# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import signal
import subprocess

from oslo_log import log as logging

import magnum_cluster_api.privsep
from magnum_cluster_api import conf

CONF = conf.CONF
LOG = logging.getLogger(__name__)


@magnum_cluster_api.privsep.haproxy_pctxt.entrypoint
def start(config_file):
    proc = subprocess.Popen(["haproxy", "-f", config_file])

    try:
        retcode = proc.wait(timeout=5)
        if retcode != 0:
            LOG.error("HAproxy failed to start")
    except subprocess.TimeoutExpired:
        LOG.info("HAproxy started successfully")

    return proc.pid


@magnum_cluster_api.privsep.haproxy_pctxt.entrypoint
def reload():
    """Reload HAproxy configuration"""

    with open(CONF.proxy.haproxy_pid_path, "r") as fd:
        pid = int(fd.read().strip())

    os.kill(pid, signal.SIGUSR2)



# Copyright (c) 2024 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

import pykube
import pytest
import responses
from magnum.objects import fields  # type: ignore
from magnum.tests.unit.objects import utils  # type: ignore
from oslo_serialization import base64, jsonutils  # type: ignore
from oslo_utils import uuidutils  # type: ignore
from responses import matchers

from magnum_cluster_api import objects, resources


@pytest.mark.parametrize(
    "auto_scaling_enabled", [True, False], ids=lambda x: f"auto_scaling_enabled={x}"
)
@pytest.mark.parametrize(
    "auto_healing_enabled", [True, False], ids=lambda x: f"auto_healing_enabled={x}"
)
class TestDriver:
    @pytest.fixture(autouse=True)
    def setup(
        self,
        context,
        mocker,
        auto_scaling_enabled,
        auto_healing_enabled,
        cluster_obj,
    ):
        self.cluster = cluster_obj

        if auto_scaling_enabled is not None:
            self.cluster.labels["auto_scaling_enabled"] = str(auto_scaling_enabled)

        if auto_healing_enabled is not None:
            self.cluster.labels["auto_healing_enabled"] = str(auto_healing_enabled)

        self.node_group = utils.get_test_nodegroup(context, labels={})
        if auto_scaling_enabled is not None:
            self.node_group.min_node_count = 1
            self.node_group.max_node_count = 3
        self.node_group.save = mocker.MagicMock()

        self.server_side_apply_matcher = matchers.query_param_matcher(
            {
                "fieldManager": "atmosphere-operator",
                "force": "True",
            }
        )

        mocker.patch(
            "magnum_cluster_api.utils.validate_flavor_name",
            return_value=True,
        )

        mocker.patch(
            "magnum_cluster_api.integrations.cinder.get_default_boot_volume_type",
            return_value="nvme",
        )

        mocker.patch(
            "magnum_cluster_api.utils.get_image_uuid",
            return_value=uuidutils.generate_uuid(),
        )

    def _assert_node_group_status(self, expected_status):
        assert self.node_group.status == expected_status
        self.node_group.save.assert_called_once()

        assert self.cluster.status == fields.ClusterStatus.UPDATE_IN_PROGRESS
        self.cluster.save.assert_called_once()

    def _response_for_cluster_with_machine_deployments(
        self, *machine_deployments, method: str = responses.GET
    ):
        obj = {
            "metadata": {
                "name": self.cluster.stack_id,
                "namespace": "magnum-system",
                "managedFields": None,
            },
            "spec": {
                "topology": {
                    "workers": {
                        "machineDeployments": list(machine_deployments),
                    }
                }
            },
        }

        match = []
        if method == responses.PATCH:
            match.append(self.server_side_apply_matcher)
            match.append(matchers.json_params_matcher(obj))

        return responses.Response(
            method,
            re.compile(
                f"http://localhost/apis/{objects.Cluster.version}/namespaces/magnum-system/{objects.Cluster.endpoint}/\\w"  # noqa
            ),
            json=obj,
            match=match,
        )

    def _response_for_machine_deployment_spec(
        self, machine_deployment_spec=None, deleted=False
    ):
        json = {"items": []}

        if machine_deployment_spec and not deleted:
            json["items"].append(
                {
                    "metadata": {
                        "name": machine_deployment_spec["name"],
                        "namespace": "magnum-system",
                    },
                    "spec": {
                        "replicas": machine_deployment_spec["replicas"],
                        "template": {
                            "metadata": machine_deployment_spec["metadata"],
                        },
                    },
                }
            )

        return responses.Response(
            responses.GET,
            "http://localhost/apis/%s/namespaces/%s/%s"
            % (
                objects.MachineDeployment.version,
                "magnum-system",
                objects.MachineDeployment.endpoint,
            ),
            match=[
                matchers.query_param_matcher(
                    {
                        "labelSelector": "cluster.x-k8s.io/cluster-name=%s,topology.cluster.x-k8s.io/deployment-name=%s"
                        % (self.cluster.stack_id, machine_deployment_spec["name"])
                    }
                ),
            ],
            json=json,
        )

    def test_create_cluster(
        self,
        requests_mock,
        context,
        ubuntu_driver,
        mock_validate_cluster,
        mock_osc,
        mock_certificates,
    ):
        with requests_mock as rsps:
            rsps.add(
                responses.GET,
                re.compile(
                    f"http://localhost/apis/{objects.Cluster.version}/namespaces/magnum-system/{objects.Cluster.endpoint}/\\w+"  # noqa
                ),
                status=404,
            )
            rsps.add_callback(
                responses.PATCH,
                f"http://localhost/api/{pykube.Namespace.version}/{pykube.Namespace.endpoint}/magnum-system",
                match=[self.server_side_apply_matcher],
                callback=lambda request: (200, {}, request.body),
            )
            rsps.add_callback(
                responses.PATCH,
                re.compile(
                    f"http://localhost/api/{pykube.Secret.version}/namespaces/magnum-system/\\w"
                ),
                match=[self.server_side_apply_matcher],
                callback=lambda request: (200, {}, request.body),
            )
            rsps.add_callback(
                responses.PATCH,
                re.compile(
                    f"http://localhost/apis/{objects.KubeadmControlPlaneTemplate.version}/namespaces/magnum-system/{objects.KubeadmControlPlaneTemplate.endpoint}/\\w+"  # noqa
                ),
                match=[self.server_side_apply_matcher],
                callback=lambda request: (200, {}, request.body),
            )
            rsps.add_callback(
                responses.PATCH,
                re.compile(
                    f"http://localhost/apis/{objects.KubeadmConfigTemplate.version}/namespaces/magnum-system/{objects.KubeadmConfigTemplate.endpoint}/\\w+"  # noqa
                ),
                match=[self.server_side_apply_matcher],
                callback=lambda request: (200, {}, request.body),
            )
            rsps.add_callback(
                responses.PATCH,
                re.compile(
                    f"http://localhost/apis/{objects.OpenStackMachineTemplate.version}/namespaces/magnum-system/{objects.OpenStackMachineTemplate.endpoint}/\\w+"  # noqa
                ),
                match=[self.server_side_apply_matcher],
                callback=lambda request: (200, {}, request.body),
            )
            rsps.add_callback(
                responses.PATCH,
                re.compile(
                    f"http://localhost/apis/{objects.OpenStackClusterTemplate.version}/namespaces/magnum-system/{objects.OpenStackClusterTemplate.endpoint}/\\w+"  # noqa
                ),
                match=[self.server_side_apply_matcher],
                callback=lambda request: (200, {}, request.body),
            )
            rsps.add_callback(
                responses.PATCH,
                re.compile(
                    f"http://localhost/apis/{objects.ClusterClass.version}/namespaces/magnum-system/{objects.ClusterClass.endpoint}/\\w+"  # noqa
                ),
                match=[self.server_side_apply_matcher],
                callback=lambda request: (200, {}, request.body),
            )
            rsps.add(
                responses.GET,
                re.compile(
                    f"http://localhost/api/{pykube.Secret.version}/namespaces/magnum-system/\\w"
                ),
                json={
                    "data": {
                        "clouds.yaml": base64.encode_as_text(
                            jsonutils.dumps(
                                {
                                    "clouds": {
                                        "default": {
                                            "region_name": "RegionOne",
                                            "verify": True,
                                            "auth": {
                                                "application_credential_id": "fake_application_credential_id",
                                                "application_credential_secret": "fake_application_credential_secret",
                                            },
                                        }
                                    }
                                }
                            )
                        ),
                    }
                },
            )
            rsps.add_callback(
                responses.PATCH,
                re.compile(
                    f"http://localhost/apis/{objects.ClusterResourceSet.version}/namespaces/magnum-system/{objects.ClusterResourceSet.endpoint}/\\w+"  # noqa
                ),
                match=[self.server_side_apply_matcher],
                callback=lambda request: (200, {}, request.body),
            )
            rsps.add_callback(
                responses.PATCH,
                re.compile(
                    f"http://localhost/apis/{objects.Cluster.version}/namespaces/magnum-system/{objects.Cluster.endpoint}/\\w+"  # noqa
                ),
                match=[self.server_side_apply_matcher],
                callback=lambda request: (200, {}, request.body),
            )

            ubuntu_driver.create_cluster(context, self.cluster, 60)

        assert self.cluster.status == fields.ClusterStatus.CREATE_IN_PROGRESS
        self.cluster.save.assert_called_once()

    def setup_node_group_tests(self, rsps, before, after):
        rsps.add(
            self._response_for_cluster_with_machine_deployments(*before),
        )
        rsps.add(
            self._response_for_cluster_with_machine_deployments(
                *after,
                method=responses.PATCH,
            )
        )

    def test_create_nodegroup(self, context, ubuntu_driver, requests_mock):
        self.cluster.status = fields.ClusterStatus.UPDATE_IN_PROGRESS
        self.node_group.status = fields.ClusterStatus.CREATE_IN_PROGRESS

        with requests_mock as rsps:
            self.setup_node_group_tests(
                rsps,
                before=[],
                after=[
                    resources.mutate_machine_deployment(
                        context,
                        self.cluster,
                        self.node_group,
                    ),
                ],
            )

            ubuntu_driver.create_nodegroup(context, self.cluster, self.node_group)

    def test_update_nodegroup(self, context, ubuntu_driver, requests_mock):
        with requests_mock as rsps:
            self.setup_node_group_tests(
                rsps,
                before=[
                    {
                        "name": self.node_group.name,
                    }
                ],
                after=[
                    resources.mutate_machine_deployment(
                        context,
                        self.cluster,
                        self.node_group,
                        {
                            "name": self.node_group.name,
                        },
                    ),
                ],
            )

            ubuntu_driver.update_nodegroup(context, self.cluster, self.node_group)

        self._assert_node_group_status(fields.ClusterStatus.UPDATE_IN_PROGRESS)

    def test_update_nodegroup_with_multiple_node_groups(
        self, context, ubuntu_driver, requests_mock
    ):
        with requests_mock as rsps:
            self.setup_node_group_tests(
                rsps,
                before=[
                    {
                        "name": "unrelated-machine-deployment",
                        "replicas": 1,
                        "metadata": {
                            "annotations": {},
                        },
                    },
                    {
                        "name": self.node_group.name,
                        "replicas": 1,
                        "metadata": {
                            "annotations": {},
                        },
                    },
                ],
                after=[
                    {
                        "name": "unrelated-machine-deployment",
                        "replicas": 1,
                        "metadata": {
                            "annotations": {},
                        },
                    },
                    resources.mutate_machine_deployment(
                        context,
                        self.cluster,
                        self.node_group,
                        {
                            "name": self.node_group.name,
                        },
                    ),
                ],
            )

            ubuntu_driver.update_nodegroup(context, self.cluster, self.node_group)

        self._assert_node_group_status(fields.ClusterStatus.UPDATE_IN_PROGRESS)

    def test_delete_nodegroup_with_multiple_node_groups(
        self, context, ubuntu_driver, requests_mock
    ):
        self.cluster.status = fields.ClusterStatus.UPDATE_IN_PROGRESS
        self.node_group.status = fields.ClusterStatus.DELETE_IN_PROGRESS

        with requests_mock as rsps:
            self.setup_node_group_tests(
                rsps,
                before=[
                    {
                        "name": "unrelated-machine-deployment",
                        "replicas": 1,
                        "metadata": {
                            "annotations": {},
                        },
                    },
                    {
                        "name": self.node_group.name,
                        "replicas": 1,
                        "metadata": {
                            "annotations": {},
                        },
                    },
                ],
                after=[
                    {
                        "name": "unrelated-machine-deployment",
                        "replicas": 1,
                        "metadata": {
                            "annotations": {},
                        },
                    },
                ],
            )

            ubuntu_driver.delete_nodegroup(context, self.cluster, self.node_group)

        assert self.node_group.status == fields.ClusterStatus.DELETE_IN_PROGRESS
        self.node_group.save.assert_called_once()



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools

import yaml
from oslo_concurrency import processutils
from oslo_log import log as logging

from magnum_cluster_api import exceptions, image_utils

LOG = logging.getLogger(__name__)


class Command:
    def __call__(self, *args, **kwargs):
        return processutils.execute("helm", *self.COMMAND, *args, **kwargs)


class VersionCommand(Command):
    COMMAND = ["version"]


class ReleaseCommand(Command):
    def __init__(self, namespace, release_name):
        self.namespace = namespace
        self.release_name = release_name

    def __call__(self, *args, **kwargs):
        return super().__call__(
            "--namespace",
            self.namespace,
            self.release_name,
            *args,
            **kwargs,
        )


class GetValuesReleaseCommand(ReleaseCommand):
    COMMAND = ["get", "values"]

    def __call__(self):
        try:
            return super().__call__(
                "--output",
                "yaml",
            )
        except processutils.ProcessExecutionError as e:
            if "release: not found" in e.stderr:
                raise exceptions.HelmReleaseNotFound(self.release_name)
            else:
                raise


class UpgradeReleaseCommand(ReleaseCommand):
    COMMAND = ["upgrade"]

    def __init__(self, namespace, release_name, chart_ref, values={}):
        super().__init__(namespace, release_name)
        self.chart_ref = chart_ref
        self.values = values

    def __call__(self):
        try:
            return super().__call__(
                self.chart_ref,
                "--install",
                "--values",
                "-",
                process_input=yaml.dump(self.values),
            )
        except processutils.ProcessExecutionError as e:
            if "UPGRADE FAILED: another operation" in e.stderr:
                LOG.info("Helm release %s is already in progress", self.release_name)
            elif "UPGRADE FAILED: release: already exists" in e.stderr:
                LOG.info("Helm release %s already exists", self.release_name)
            else:
                raise


class DeleteReleaseCommand(ReleaseCommand):
    COMMAND = ["delete"]

    def __init__(self, namespace, release_name, skip_missing=False):
        super().__init__(namespace, release_name)
        self.skip_missing = skip_missing

    def __call__(self):
        try:
            return super().__call__()
        except processutils.ProcessExecutionError as e:
            if "release: not found" in e.stderr:
                if self.skip_missing:
                    pass
                else:
                    raise exceptions.HelmReleaseNotFound(self.release_name)
            else:
                raise


class TemplateReleaseCommand(ReleaseCommand):
    COMMAND = ["template"]

    def __init__(self, namespace, release_name, chart_ref, values={}):
        super().__init__(namespace, release_name)
        self.chart_ref = chart_ref
        self.values = values

    def __call__(self, repository=None, replacements=[]):
        try:
            docs = []
            data, _ = super().__call__(
                self.chart_ref,
                "--values",
                "-",
                process_input=yaml.dump(self.values),
            )

            # NOTE(oleks): ClusterResourceSet fails to be applied without namespace set in resources.
            #              On the other hand, helm template doesn't output namespace. We set it manually
            #              until  https://github.com/helm/helm/issues/10737 is fixed.
            for doc in yaml.safe_load_all(data):
                if doc["kind"] in ("DaemonSet", "Deployment", "StatefulSet"):
                    for container in itertools.chain(
                        doc["spec"]["template"]["spec"].get("initContainers", []),
                        doc["spec"]["template"]["spec"]["containers"],
                    ):
                        for src, dst in replacements:
                            container["image"] = container["image"].replace(src, dst)
                        if repository:
                            container["image"] = image_utils.get_image(
                                container["image"], repository
                            )

                if doc["kind"] not in ("ClusterRole", "ClusterRoleBinding"):
                    doc["metadata"]["namespace"] = self.namespace
                docs.append(doc)
            return yaml.safe_dump_all(docs, default_flow_style=False)

        except processutils.ProcessExecutionError:
            LOG.info("Helm template %s failed", self.release_name)
            raise



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import pykube
from magnum.common import clients, exception
from manilaclient.v2 import client as manilaclient


class OpenStackClients(clients.OpenStackClients):
    """Convenience class to create and cache client instances."""

    def __init__(self, context):
        super(OpenStackClients, self).__init__(context)
        self._manila = None

    @exception.wrap_keystone_exception
    def manila(self):
        if self._manila:
            return self._manila
        endpoint_type = self._get_client_option("manila", "endpoint_type")
        region_name = self._get_client_option("manila", "region_name")
        manilaclient_version = self._get_client_option("manila", "api_version")
        endpoint = self.url_for(
            service_type="sharev2", interface=endpoint_type, region_name=region_name
        )
        args = {
            "cacert": self._get_client_option("manila", "ca_file"),
            "insecure": self._get_client_option("manila", "insecure"),
        }

        session = self.keystone().session
        self._manila = manilaclient.Client(
            manilaclient_version, session=session, service_catalog_url=endpoint, **args
        )
        return self._manila


def get_pykube_api() -> pykube.HTTPClient:
    return pykube.HTTPClient(pykube.KubeConfig.from_env())


def get_openstack_api(context) -> OpenStackClients:
    return OpenStackClients(context)



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import json
import re
import string
import textwrap

import pykube
import shortuuid
import yaml
from magnum import objects as magnum_objects
from magnum.api import attr_validator
from magnum.common import context, exception, neutron, octavia
from magnum.common import utils as magnum_utils
from oslo_config import cfg
from oslo_serialization import base64
from oslo_utils import strutils, uuidutils
from tenacity import retry, retry_if_exception_type

from magnum_cluster_api import clients
from magnum_cluster_api import exceptions as mcapi_exceptions
from magnum_cluster_api import image_utils, images, objects

AVAILABLE_OPERATING_SYSTEMS = ["ubuntu", "flatcar", "rockylinux"]
CONF = cfg.CONF


def get_cluster_api_cloud_config_secret_name(cluster: magnum_objects.Cluster) -> str:
    return f"{cluster.stack_id}-cloud-config"


def get_or_generate_cluster_api_cloud_config_secret_name(
    api: pykube.HTTPClient, cluster: magnum_objects.Cluster
) -> str:
    return f"{get_or_generate_cluster_api_name(api, cluster)}-cloud-config"


def get_or_generate_cluster_api_name(
    api: pykube.HTTPClient, cluster: magnum_objects.Cluster
) -> str:
    if cluster.stack_id is None:
        cluster.stack_id = generate_cluster_api_name(api)
        cluster.save()
    return cluster.stack_id


@retry(retry=retry_if_exception_type(exception.Conflict))
def generate_cluster_api_name(
    api: pykube.HTTPClient,
) -> str:
    alphabet = string.ascii_lowercase + string.digits
    su = shortuuid.ShortUUID(alphabet=alphabet)

    name = "kube-%s" % (su.random(length=5))
    if cluster_exists(api, name):
        raise exception.Conflict("Generated name already exists")
    return name


def cluster_exists(api: pykube.HTTPClient, name: str) -> bool:
    try:
        objects.Cluster.objects(api, namespace="magnum-system").get(name=name)
        return True
    except pykube.exceptions.ObjectDoesNotExist:
        return False


def get_cloud_ca_cert() -> str:
    """
    Get cloud ca certificate.
    """
    return magnum_utils.get_openstack_ca()


def get_capi_client_ca_cert() -> str:
    ca_file = CONF.capi_client.ca_file

    if ca_file:
        with open(ca_file) as fd:
            return fd.read()
    else:
        return ""


def generate_cloud_controller_manager_config(
    ctx: context.RequestContext,
    api: pykube.HTTPClient,
    cluster: magnum_objects.Cluster,
) -> str:
    """
    Generate coniguration for openstack-cloud-controller-manager if it does
    already exist.
    """

    osc = clients.get_openstack_api(ctx)
    data = pykube.Secret.objects(api, namespace="magnum-system").get_by_name(
        get_cluster_api_cloud_config_secret_name(cluster)
    )
    clouds_yaml = base64.decode_as_text(data.obj["data"]["clouds.yaml"])
    cloud_config = yaml.safe_load(clouds_yaml)

    octavia_provider = cluster.labels.get("octavia_provider", "amphora")
    octavia_lb_algorithm = cluster.labels.get("octavia_lb_algorithm")

    if octavia_provider == "amphora" and octavia_lb_algorithm is None:
        octavia_lb_algorithm = "ROUND_ROBIN"
    elif octavia_provider == "ovn" and octavia_lb_algorithm is None:
        octavia_lb_algorithm = "SOURCE_IP_PORT"
    elif octavia_provider == "ovn" and octavia_lb_algorithm != "SOURCE_IP_PORT":
        raise mcapi_exceptions.InvalidOctaviaLoadBalancerAlgorithm(
            octavia_lb_algorithm=octavia_lb_algorithm
        )

    return textwrap.dedent(
        f"""\
        [Global]
        auth-url={osc.url_for(service_type="identity", interface="public")}
        region={cloud_config["clouds"]["default"]["region_name"]}
        application-credential-id={cloud_config["clouds"]["default"]["auth"]["application_credential_id"]}
        application-credential-secret={cloud_config["clouds"]["default"]["auth"]["application_credential_secret"]}
        tls-insecure={"false" if CONF.drivers.verify_ca else "true"}
        {"ca-file=/etc/config/ca.crt" if get_cloud_ca_cert() else ""}
        [LoadBalancer]
        lb-provider={octavia_provider}
        lb-method={octavia_lb_algorithm}
        """
    )


def generate_manila_csi_cloud_config(
    ctx: context.RequestContext,
    api: pykube.HTTPClient,
    cluster: magnum_objects.Cluster,
) -> dict[str, str]:
    """
    Generate coniguration of Openstack authentication  for manila csi
    """
    osc = clients.get_openstack_api(ctx)
    data = pykube.Secret.objects(api, namespace="magnum-system").get_by_name(
        get_cluster_api_cloud_config_secret_name(cluster)
    )
    clouds_yaml = base64.decode_as_text(data.obj["data"]["clouds.yaml"])
    cloud_config = yaml.safe_load(clouds_yaml)

    config = {
        "os-authURL": osc.url_for(service_type="identity", interface="public"),
        "os-region": cloud_config["clouds"]["default"]["region_name"],
        "os-applicationCredentialID": cloud_config["clouds"]["default"]["auth"][
            "application_credential_id"
        ],
        "os-applicationCredentialSecret": cloud_config["clouds"]["default"]["auth"][
            "application_credential_secret"
        ],
        "os-TLSInsecure": (
            ("false" if CONF.drivers.verify_ca else "true")
            if cloud_config["clouds"]["default"]["verify"]
            else "true"
        ),
    }

    if get_cloud_ca_cert():
        config["os-certAuthorityPath"] = "/etc/config/ca.crt"

    return config


def get_kube_tag(cluster: magnum_objects.Cluster) -> str:
    return cluster.labels.get("kube_tag", "v1.25.3")


def get_auto_scaling_enabled(cluster: magnum_objects.Cluster) -> bool:
    return get_cluster_label_as_bool(cluster, "auto_scaling_enabled", False)


def get_auto_healing_enabled(cluster: magnum_objects.Cluster) -> bool:
    return get_cluster_label_as_bool(cluster, "auto_healing_enabled", True)


def get_cluster_container_infra_prefix(cluster: magnum_objects.Cluster) -> str:
    return cluster.labels.get("container_infra_prefix", "")


def get_cluster_floating_ip_disabled(cluster: magnum_objects.Cluster) -> bool:
    return not get_cluster_label_as_bool(cluster, "master_lb_floating_ip_enabled", True)


def generate_containerd_config(
    cluster: magnum_objects.Cluster,
):
    image_repository = get_cluster_container_infra_prefix(cluster)
    sandbox_image = image_utils.get_image(images.PAUSE, image_repository)

    return textwrap.dedent(
        """\
        # Use config version 2 to enable new configuration fields.
        # Config file is parsed as version 1 by default.
        version = 2

        imports = ["/etc/containerd/conf.d/*.toml"]

        [plugins]
        [plugins."io.containerd.grpc.v1.cri"]
            sandbox_image = "{sandbox_image}"
        [plugins."io.containerd.grpc.v1.cri".containerd.runtimes.runc]
            runtime_type = "io.containerd.runc.v2"
        [plugins."io.containerd.grpc.v1.cri".containerd.runtimes.runc.options]
            SystemdCgroup = true
        """
    ).format(sandbox_image=sandbox_image)


def generate_systemd_proxy_config(cluster: magnum_objects.Cluster):
    if (
        cluster.cluster_template.http_proxy is not None
        or cluster.cluster_template.https_proxy is not None
    ):
        return textwrap.dedent(
            """\
            [Service]
            Environment="http_proxy={http_proxy}"
            Environment="HTTP_PROXY={http_proxy}"
            Environment="https_proxy={https_proxy}"
            Environment="HTTPS_PROXY={https_proxy}"
            Environment="no_proxy={no_proxy}"
            Environment="NO_PROXY={no_proxy}"
            """
        ).format(
            http_proxy=cluster.cluster_template.http_proxy,
            https_proxy=cluster.cluster_template.https_proxy,
            no_proxy=cluster.cluster_template.no_proxy,
        )
    else:
        return ""


def generate_apt_proxy_config(cluster: magnum_objects.Cluster):
    if (
        cluster.cluster_template.http_proxy is not None
        or cluster.cluster_template.https_proxy is not None
    ):
        return textwrap.dedent(
            """\
            Acquire::http::Proxy "{http_proxy}";
            Acquire::https::Proxy "{https_proxy}";
            """
        ).format(
            http_proxy=cluster.cluster_template.http_proxy,
            https_proxy=cluster.cluster_template.https_proxy,
        )
    else:
        return ""


def get_node_group_min_node_count(
    node_group: magnum_objects.NodeGroup,
    default=1,
) -> int:
    if node_group.min_node_count == 0:
        return default
    return node_group.min_node_count


def get_node_group_max_node_count(
    node_group: magnum_objects.NodeGroup,
) -> int:
    if node_group.max_node_count is None:
        return get_node_group_label_as_int(
            node_group,
            "max_node_count",
            get_node_group_min_node_count(node_group) + 1,
        )
    return node_group.max_node_count


def get_node_group_label_as_int(
    node_group: magnum_objects.NodeGroup,
    key: str,
    default: int,
) -> int:
    value = node_group.labels.get(key, str(default))
    return strutils.validate_integer(value, key)


def get_cluster_label_as_int(
    cluster: magnum_objects.Cluster, key: str, default: int
) -> int:
    value = cluster.labels.get(key, default)
    return strutils.validate_integer(value, key)


def get_cluster_label_as_bool(
    cluster: magnum_objects.Cluster, key: str, default: bool
) -> bool:
    value = cluster.labels.get(key, default)
    return strutils.bool_from_string(value, strict=True)


def delete_loadbalancers(ctx, cluster):
    # NOTE(mnaser): This code is duplicated from magnum.common.octavia
    #               since the original code is very Heat-specific.
    pattern = r"Kubernetes .+ from cluster %s" % cluster.uuid

    admin_ctx = context.get_admin_context()
    admin_clients = clients.get_openstack_api(admin_ctx)
    user_clients = clients.get_openstack_api(ctx)

    candidates = set()

    try:
        octavia_admin_client = admin_clients.octavia()
        octavia_client = user_clients.octavia()

        # Get load balancers created for service/ingress
        lbs = octavia_client.load_balancer_list().get("loadbalancers", [])
        lbs = [lb for lb in lbs if re.match(pattern, lb["description"])]
        deleted = octavia._delete_loadbalancers(
            ctx, lbs, cluster, octavia_admin_client, remove_fip=True
        )
        candidates.update(deleted)

        if not candidates:
            return

        octavia.wait_for_lb_deleted(octavia_client, candidates)
    except Exception as e:
        raise exception.PreDeletionFailed(cluster_uuid=cluster.uuid, msg=str(e))


def format_event_message(event: pykube.Event):
    return "%s: %s" % (
        event.obj["reason"],
        event.obj["message"],
    )


def validate_flavor_name(cli: clients.OpenStackClients, flavor: str):
    """Check if a flavor with this specified name exists"""

    if flavor is None:
        return
    flavor_list = cli.nova().flavors.list()
    for f in flavor_list:
        if f.name == flavor:
            return
        if f.id == flavor:
            raise mcapi_exceptions.OpenstackFlavorInvalidName(flavor=flavor)
    raise exception.FlavorNotFound(flavor=flavor)


def validate_cluster(ctx: context.RequestContext, cluster: magnum_objects.Cluster):
    # Check master count
    if (cluster.master_count % 2) == 0:
        raise mcapi_exceptions.ClusterMasterCountEven
    # Validate flavors
    osc = clients.get_openstack_api(ctx)
    validate_flavor_name(osc, cluster.master_flavor_id)
    validate_flavor_name(osc, cluster.flavor_id)

    # Check if fixed_network exists
    if cluster.fixed_network:
        if uuidutils.is_uuid_like(cluster.fixed_network):
            neutron.get_network(
                ctx,
                cluster.fixed_network,
                source="id",
                target="name",
                external=False,
            )
        else:
            neutron.get_network(
                ctx,
                cluster.fixed_network,
                source="name",
                target="id",
                external=False,
            )

    # Check if fixed_subnet exists
    if cluster.fixed_subnet:
        if uuidutils.is_uuid_like(cluster.fixed_subnet):
            neutron.get_subnet(ctx, cluster.fixed_subnet, source="id", target="name")
        else:
            neutron.get_subnet(ctx, cluster.fixed_subnet, source="name", target="id")


def validate_nodegroup(
    nodegroup: magnum_objects.NodeGroup, ctx: context.RequestContext
):
    # Validate flavors
    osc = clients.get_openstack_api(ctx)
    validate_flavor_name(osc, nodegroup.flavor_id)


def get_operating_system(cluster: magnum_objects.Cluster):
    cluster_distro = cluster.cluster_template.cluster_distro
    for ops in AVAILABLE_OPERATING_SYSTEMS:
        if cluster_distro.startswith(ops):
            return ops
    return None


def get_image_uuid(image_ref: str, ctx: context.RequestContext):
    """Get image uuid from image ref

    :param image_ref: Image id or name
    """
    osc = clients.get_openstack_api(ctx)
    image_obj = attr_validator.validate_image(osc, image_ref)
    return image_obj.get("id")


def convert_to_rfc1123(input: str) -> str:
    """
    Convert a given string to RFC1123 format.

    :param input: The string to be converted.
    :type input: str

    :return: The converted string in RFC1123 format.
    :rtype: str
    """
    return re.sub(r"[^a-zA-Z0-9]+", "-", input).lower()


def get_keystone_auth_default_policy(cluster: magnum_objects.Cluster):
    default_policy = [
        {
            "resource": {
                "verbs": ["list"],
                "resources": ["pods", "services", "deployments", "pvc"],
                "version": "*",
                "namespace": "default",
            },
            "match": [
                {"type": "role", "values": ["member"]},
                {"type": "project", "values": [cluster.project_id]},
            ],
        }
    ]

    try:
        with open(CONF.kubernetes.keystone_auth_default_policy) as f:
            return json.loads(f.read())
    except Exception:
        return default_policy


def kube_apply_patch(resource):
    if "metadata" in resource.obj:
        resource.obj["metadata"]["managedFields"] = None

    resp = resource.api.patch(
        **resource.api_kwargs(
            headers={
                "Content-Type": "application/apply-patch+yaml",
            },
            params={
                "fieldManager": "atmosphere-operator",
                "force": True,
            },
            data=json.dumps(resource.obj),
        )
    )

    resource.api.raise_for_status(resp)
    resource.set_obj(resp.json())



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log as logging

from magnum_cluster_api import service
from magnum_cluster_api.proxy import manager

CONF = cfg.CONF


def main():
    logging.register_options(CONF)
    logging.setup(CONF, "magnum-cluster-api-proxy")

    server = service.Service(manager=manager.ProxyManager)
    service.serve(server)
    service.wait()



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import configparser
import textwrap

import pytest
from oslo_serialization import base64

from magnum_cluster_api import exceptions, objects


@pytest.fixture
def mock_openstack_cluster(mocker):
    mock_api = mocker.Mock()
    mocker.patch.object(
        objects.OpenStackCluster,
        "identity_ref_secret",
        new_callable=mocker.PropertyMock,
        return_value=mocker.Mock(
            obj={
                "apiVersion": "v1",
                "kind": "Secret",
                "metadata": {
                    "name": "test-identity",
                    "namespace": "test-namespace",
                },
                "data": {
                    "clouds.yaml": base64.encode_as_text(
                        textwrap.dedent(
                            """
                                clouds:
                                  default:
                                    region_name: RegionOne
                                    verify: false
                                    auth:
                                      auth_url: https://example.com:5000/v3
                                      application_credential_id: cluster-api
                                      application_credential_secret: secret123
                                """
                        )
                    )
                },
            }
        ),
    )

    return objects.OpenStackCluster(
        mock_api,
        {
            "apiVersion": objects.OpenStackCluster.version,
            "kind": objects.OpenStackCluster.kind,
            "metadata": {
                "name": "test-cluster",
                "namespace": "test-namespace",
            },
            "spec": {
                "identityRef": {
                    "name": "test-identity",
                }
            },
            "status": {
                "externalNetwork": {
                    "id": "foo",
                },
                "network": {
                    "id": "bar",
                    "subnet": {
                        "id": "baz",
                    },
                },
            },
        },
    )


class TestOpenStackCluster:
    def test_floating_network_id(self, mock_openstack_cluster):
        assert (
            mock_openstack_cluster.floating_network_id
            == mock_openstack_cluster.obj["status"]["externalNetwork"]["id"]
        )

    def test_floating_network_id_not_ready(self, mock_openstack_cluster):
        del mock_openstack_cluster.obj["status"]["externalNetwork"]
        with pytest.raises(exceptions.OpenStackClusterExternalNetworkNotReady):
            mock_openstack_cluster.floating_network_id

    def test_network_id(self, mock_openstack_cluster):
        assert (
            mock_openstack_cluster.network_id
            == mock_openstack_cluster.obj["status"]["network"]["id"]
        )

    def test_network_id_not_ready(self, mock_openstack_cluster):
        del mock_openstack_cluster.obj["status"]["network"]["id"]
        with pytest.raises(exceptions.OpenStackClusterNetworkNotReady):
            mock_openstack_cluster.network_id

    def test_subnet_id(self, mock_openstack_cluster):
        assert (
            mock_openstack_cluster.subnet_id
            == mock_openstack_cluster.obj["status"]["network"]["subnet"]["id"]
        )

    def test_subnet_id_not_ready(self, mock_openstack_cluster):
        del mock_openstack_cluster.obj["status"]["network"]["subnet"]["id"]
        with pytest.raises(exceptions.OpenStackClusterSubnetNotReady):
            mock_openstack_cluster.subnet_id

    def test_cloud_controller_manager_config(self, mock_openstack_cluster):
        config = configparser.ConfigParser()
        config.read_string(mock_openstack_cluster.cloud_controller_manager_config)

        assert {s: dict(config.items(s)) for s in config.sections()} == {
            "Global": {
                "auth-url": mock_openstack_cluster.cloud_config["auth"]["auth_url"],
                "region": mock_openstack_cluster.cloud_config["region_name"],
                "application-credential-id": mock_openstack_cluster.cloud_config[
                    "auth"
                ]["application_credential_id"],
                "application-credential-secret": mock_openstack_cluster.cloud_config[
                    "auth"
                ]["application_credential_secret"],
                "tls-insecure": (
                    "false" if mock_openstack_cluster.cloud_config["verify"] else "true"
                ),
            },
            "LoadBalancer": {
                "floating-network-id": mock_openstack_cluster.floating_network_id,
                "network-id": mock_openstack_cluster.network_id,
                "subnet-id": mock_openstack_cluster.subnet_id,
            },
        }



# Copyright (c) 2024 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import textwrap
from dataclasses import dataclass, field

import yaml

from magnum_cluster_api import objects


@dataclass
class DiskConfig:
    type: str
    mount_path: str


@dataclass
class Volumes:
    control_plane_disks: list[DiskConfig] = field(default_factory=list)
    worker_disks: list[DiskConfig] = field(default_factory=list)

    def _disk_setup(self, disks: list[DiskConfig]) -> dict:
        return {
            "partitions": [
                {
                    "device": f"/dev/vd{chr(ord('a') + index + 1)}",
                    "tableType": "gpt",
                    "layout": True,
                    "overwrite": False,
                }
                for index in range(len(disks))
            ],
            "filesystems": [
                {
                    "label": f"{disk.type}_disk",
                    "filesystem": "ext4",
                    "device": f"/dev/vd{chr(ord('a') + index + 1)}",
                    "extraOpts": [
                        "-F",
                        "-E",
                        "lazy_itable_init=1,lazy_journal_init=1",
                    ],
                }
                for index, disk in enumerate(disks)
            ],
        }

    def _mounts(self, disks: list[DiskConfig]) -> list[list[str]]:
        return [
            [
                f"LABEL={disk.type}_disk",
                disk.mount_path,
            ]
            for disk in disks
        ]

    def _additional_block_devices(self, disks: list[DiskConfig]) -> str:
        return "\n".join(
            [
                textwrap.dedent(
                    f"""\
                - name: {disk.type}
                  sizeGiB: {{{{ .{disk.type}VolumeSize }}}}
                  storage:
                    type: Volume
                    volume:
                      type: "{{{{ .{disk.type}VolumeType }}}}"
                      availabilityZone: "{{{{ .availabilityZone }}}}"
                """
                )
                for disk in disks
            ]
        )

    @property
    def definitions(self) -> list[dict]:
        definitions = []

        if len(self.control_plane_disks) > 0:
            definitions += [
                {
                    "selector": {
                        "apiVersion": objects.OpenStackMachineTemplate.version,
                        "kind": objects.OpenStackMachineTemplate.kind,
                        "matchResources": {
                            "controlPlane": True,
                        },
                    },
                    "jsonPatches": [
                        {
                            "op": "add",
                            "path": "/spec/template/spec/additionalBlockDevices",
                            "valueFrom": {
                                "template": self._additional_block_devices(
                                    self.control_plane_disks
                                ),
                            },
                        },
                    ],
                },
                {
                    "selector": {
                        "apiVersion": objects.KubeadmControlPlaneTemplate.version,
                        "kind": objects.KubeadmControlPlaneTemplate.kind,
                        "matchResources": {
                            "controlPlane": True,
                        },
                    },
                    "jsonPatches": [
                        {
                            "op": "add",
                            "path": "/spec/template/spec/kubeadmConfigSpec/diskSetup",
                            "valueFrom": {
                                "template": yaml.dump(
                                    self._disk_setup(self.control_plane_disks)
                                ),
                            },
                        },
                        {
                            "op": "add",
                            "path": "/spec/template/spec/kubeadmConfigSpec/mounts",
                            "valueFrom": {
                                "template": yaml.dump(
                                    self._mounts(self.control_plane_disks)
                                ),
                            },
                        },
                    ],
                },
            ]

        if len(self.worker_disks) > 0:
            definitions += [
                {
                    "selector": {
                        "apiVersion": objects.OpenStackMachineTemplate.version,
                        "kind": objects.OpenStackMachineTemplate.kind,
                        "matchResources": {
                            "machineDeploymentClass": {
                                "names": ["default-worker"],
                            },
                        },
                    },
                    "jsonPatches": [
                        {
                            "op": "add",
                            "path": "/spec/template/spec/additionalBlockDevices",
                            "valueFrom": {
                                "template": self._additional_block_devices(
                                    self.worker_disks
                                ),
                            },
                        },
                    ],
                },
                {
                    "selector": {
                        "apiVersion": objects.KubeadmConfigTemplate.version,
                        "kind": objects.KubeadmConfigTemplate.kind,
                        "matchResources": {
                            "machineDeploymentClass": {
                                "names": ["default-worker"],
                            }
                        },
                    },
                    "jsonPatches": [
                        {
                            "op": "add",
                            "path": "/spec/template/spec/diskSetup",
                            "valueFrom": {
                                "template": yaml.dump(
                                    self._disk_setup(self.worker_disks)
                                ),
                            },
                        },
                        {
                            "op": "add",
                            "path": "/spec/template/spec/mounts",
                            "valueFrom": {
                                "template": yaml.dump(self._mounts(self.worker_disks)),
                            },
                        },
                    ],
                },
            ]

        return definitions



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import dataclasses
import os
import socket
from datetime import datetime, timezone

import haproxyadmin
from oslo_log import log as logging
from pyroute2 import netns

from magnum_cluster_api import objects
from magnum_cluster_api.proxy import utils

LOG = logging.getLogger(__name__)


@dataclasses.dataclass
class ProxiedCluster:
    """A cluster that is proxied by this service."""

    NODE_LABEL = "magnum-cluster-api.vexxhost.com/node"
    CLUSTER_LABEL = "magnum-cluster-api.vexxhost.com/proxied-cluster"
    SERVICE_LABEL = "magnum-cluster-api.vexxhost.com/proxied-service"
    TIMESTAMP_ANNOTATION = "magnum-cluster-api.vexxhost.com/timestamp"

    name: str
    internal_ip: str
    namespace: str

    @classmethod
    def from_openstack_cluster(
        self, cluster: objects.OpenStackCluster
    ) -> "ProxiedCluster":
        spec = cluster.obj.get("spec", {})

        # NOTE(mnaser): If the API server floating IP is disabled, we don't
        #               need to proxy it.
        if spec.get("disableAPIServerFloatingIP", False) is False:
            return None

        status = cluster.obj.get("status", {})

        network_id = status.get("network", {}).get("id")
        internal_ip = status.get("apiServerLoadBalancer", {}).get("internalIP")

        if network_id is None:
            LOG.debug("No network ID found for cluster %s", cluster.name)
            return

        namespaces = [n for n in netns.listnetns() if n.endswith(network_id)]
        if len(namespaces) == 0:
            LOG.debug("No namespaces found for network %s", network_id)
            return

        return ProxiedCluster(
            name=cluster.metadata["labels"]["cluster.x-k8s.io/cluster-name"],
            internal_ip=internal_ip,
            namespace=namespaces[0],
        )

    @classmethod
    def from_endpoint_slice(
        cls, endpoint_slice: objects.EndpointSlice
    ) -> "ProxiedCluster":
        """
        Returns a ProxiedCluster object from an endpoint slice.

        This is used when we're looking up a cluster from an endpoint slice
        and we need to get the cluster name from the endpoint slice.
        """
        return ProxiedCluster(
            name=endpoint_slice.metadata["labels"][cls.CLUSTER_LABEL],
            namespace=endpoint_slice.namespace,
            # TODO(mnaser): We can try and figure out a way to get this, but
            #               when we're looking up from the endpoint slice, we
            #               don't really need it.
            internal_ip="",
        )

    @property
    def endpoint_slice_name(self) -> str:
        return f"{self.name}-{socket.gethostname()}"

    @property
    def endpoint_slice_labels(self) -> dict:
        return {
            "kubernetes.io/service-name": self.name,
            self.NODE_LABEL: socket.gethostname(),
            self.CLUSTER_LABEL: self.name,
            self.SERVICE_LABEL: "true",
        }

    @property
    def endpoint_slice_annotations(self) -> dict:
        return {
            self.TIMESTAMP_ANNOTATION: datetime.now(timezone.utc).isoformat(),
        }

    @property
    def endpoint_slice_endpoints(self) -> list:
        ip = os.getenv("POD_IP", utils.get_default_ip_address())

        return [
            {
                "addresses": [ip],
                "conditions": {
                    "ready": True,
                },
                "nodeName": socket.gethostname(),
            }
        ]

    def endpoint_slice_ports(self, haproxy_port) -> list:
        return [
            {
                "name": "https",
                "port": haproxy_port,
                "protocol": "TCP",
            }
        ]

    @property
    def kubeconfig_secret_name(self) -> str:
        return f"{self.name}-kubeconfig"

    @property
    def backend_name(self) -> str:
        """
        Returns the name of the backend for this cluster.
        """
        return f"{self.name}.magnum-system"

    @property
    def backend(self) -> haproxyadmin.backend.Backend:
        """
        Returns the backend object for this cluster.
        """
        hap = haproxyadmin.haproxy.HAProxy(socket_dir="/var/run")
        return hap.backend(self.backend_name)

    @property
    def healthy(self) -> bool:
        """
        Returns whether the backend is healthy or not.
        """
        try:
            return self.backend.status == "UP"
        except ValueError:
            LOG.error("Backend %s does not exist", self.backend_name)
            return False



# Copyright (c) 2024 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import pytest
import tenacity
from magnum import objects as magnum_objects  # type: ignore
from magnum.objects import fields  # type: ignore
from magnum.tests.unit.objects import utils  # type: ignore

from magnum_cluster_api import clients, objects


class TestDriver:
    @pytest.fixture(autouse=True)
    def setup(self, cluster):
        self.api = clients.get_pykube_api()
        self.cluster = cluster

    def _assert_machine_deployment_config_matches_node_group(self, md, node_group):
        assert md is not None
        # TODO: more?

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(10),
        wait=tenacity.wait_fixed(1),
    )
    def _assert_machine_deployments_for_node_groups(
        self, *node_groups: magnum_objects.NodeGroup
    ):
        mds = objects.MachineDeployment.objects(
            self.api, namespace="magnum-system"
        ).filter(
            selector={
                "cluster.x-k8s.io/cluster-name": self.cluster.stack_id,
            },
        )

        worker_ngs = [ng for ng in node_groups if ng.role != "master"]

        assert len(mds) == len(worker_ngs), "Expected %d MachineDeployments, got %d" % (
            len(worker_ngs),
            len(mds),
        )

        # NOTE(mnaser): We need to loop over all the node groups and make sure
        #               that the machine deployments are created for them.
        for ng in worker_ngs:
            md = objects.MachineDeployment.for_node_group(self.api, self.cluster, ng)
            self._assert_machine_deployment_config_matches_node_group(md, ng)

        # NOTE(mnaser): We also need to make sure there are no extra machine
        #               deployments created.
        for md in mds:
            assert md.labels["topology.cluster.x-k8s.io/deployment-name"] in [
                ng.name for ng in node_groups
            ]

    def test_default_node_group(self):
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)

    def _create_node_group(self, context, driver, node_group_name, cluster_template):
        new_node_group = utils.get_test_nodegroup(
            context,
            name="default-worker",
            role="worker",
            node_count=1,
            flavor_id=cluster_template.master_flavor_id,
            image_id=cluster_template.image_id,
            labels={},
            status=fields.ClusterStatus.CREATE_IN_PROGRESS,
        )
        new_node_group.name = node_group_name
        new_node_group.save = mock.MagicMock()

        self.cluster.status = fields.ClusterStatus.UPDATE_IN_PROGRESS
        driver.create_nodegroup(context, self.cluster, new_node_group)

        return new_node_group

    def test_upgrade_cluster(self, context, ubuntu_driver, cluster_template):
        cluster_template.labels["kube_tag"] = "v1.26.3"

        cluster_resource = objects.Cluster.for_magnum_cluster(self.api, self.cluster)
        current_observed_generation = cluster_resource.observed_generation

        ubuntu_driver.upgrade_cluster(
            context, self.cluster, cluster_template, None, None
        )

        cluster_resource.wait_for_observed_generation_changed(
            existing_observed_generation=current_observed_generation,
        )

        cluster_resource = objects.Cluster.for_magnum_cluster(self.api, self.cluster)
        assert cluster_resource.observed_generation != current_observed_generation

        self.cluster.save.assert_not_called()

    def test_upgrade_cluster_to_same_version(
        self, kube_tag, context, ubuntu_driver, cluster_template
    ):
        cluster_template.labels["kube_tag"] = kube_tag

        cluster_resource = objects.Cluster.for_magnum_cluster(self.api, self.cluster)
        current_observed_generation = cluster_resource.observed_generation

        ubuntu_driver.upgrade_cluster(
            context, self.cluster, cluster_template, None, None
        )

        cluster_resource = objects.Cluster.for_magnum_cluster(self.api, self.cluster)
        assert cluster_resource.observed_generation == current_observed_generation

        self.cluster.save.assert_not_called()

    def test_upgrade_cluster_with_multiple_node_groups(
        self,
        mocker,
        control_plane_node_group_obj,
        worker_node_group_obj,
        mock_validate_nodegroup,
        context,
        ubuntu_driver,
        cluster_template,
    ):
        new_node_group = self._create_node_group(
            context, ubuntu_driver, "high-cpu", cluster_template
        )

        mocker.patch(
            "magnum.objects.NodeGroup.list",
            return_value=[
                control_plane_node_group_obj,
                worker_node_group_obj,
                new_node_group,
            ],
        )

        cluster_template.labels["kube_tag"] = "v1.26.3"

        cluster_resource = objects.Cluster.for_magnum_cluster(self.api, self.cluster)
        current_observed_generation = cluster_resource.observed_generation

        ubuntu_driver.upgrade_cluster(
            context, self.cluster, cluster_template, None, None
        )

        cluster_resource = objects.Cluster.for_magnum_cluster(self.api, self.cluster)
        assert cluster_resource.observed_generation != current_observed_generation

        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
        )

        self.cluster.save.assert_not_called()

    def test_create_node_group(
        self, mock_validate_nodegroup, context, ubuntu_driver, cluster_template
    ):
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)
        new_node_group = self._create_node_group(
            context, ubuntu_driver, "high-cpu", cluster_template
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            new_node_group,
        )

    def test_create_and_delete_node_group(
        self, mock_validate_nodegroup, context, ubuntu_driver, cluster_template
    ):
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)
        new_node_group = self._create_node_group(
            context, ubuntu_driver, "high-cpu", cluster_template
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            new_node_group,
        )
        ubuntu_driver.delete_nodegroup(context, self.cluster, new_node_group)
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)

    def test_create_two_node_groups(
        self, mock_validate_nodegroup, context, ubuntu_driver, cluster_template
    ):
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)
        first_new_node_group = self._create_node_group(
            context, ubuntu_driver, "high-cpu", cluster_template
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            first_new_node_group,
        )
        second_new_node_group = self._create_node_group(
            context, ubuntu_driver, "high-memory", cluster_template
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            first_new_node_group,
            second_new_node_group,
        )

    def test_create_and_delete_two_node_groups_deleting_newest_first(
        self, mock_validate_nodegroup, context, ubuntu_driver, cluster_template
    ):
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)
        first_new_node_group = self._create_node_group(
            context, ubuntu_driver, "high-cpu", cluster_template
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            first_new_node_group,
        )
        second_new_node_group = self._create_node_group(
            context, ubuntu_driver, "high-memory", cluster_template
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            first_new_node_group,
            second_new_node_group,
        )
        ubuntu_driver.delete_nodegroup(context, self.cluster, second_new_node_group)
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            first_new_node_group,
        )
        ubuntu_driver.delete_nodegroup(context, self.cluster, first_new_node_group)
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)

    def test_create_and_delete_two_node_groups_deleting_oldest_first(
        self, mock_validate_nodegroup, context, ubuntu_driver, cluster_template
    ):
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)
        first_new_node_group = self._create_node_group(
            context, ubuntu_driver, "high-cpu", cluster_template
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            first_new_node_group,
        )
        second_new_node_group = self._create_node_group(
            context, ubuntu_driver, "high-memory", cluster_template
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            first_new_node_group,
            second_new_node_group,
        )
        ubuntu_driver.delete_nodegroup(context, self.cluster, first_new_node_group)
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            second_new_node_group,
        )
        ubuntu_driver.delete_nodegroup(context, self.cluster, second_new_node_group)
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import concurrent.futures
import os
import shutil
import subprocess
import tempfile

import click
import platformdirs
import requests
from diskcache import FanoutCache

from magnum_cluster_api import image_utils

CACHE = FanoutCache(
    directory=platformdirs.user_cache_dir("magnum_cluster_api", "vexxhost"),
)

# NOTE(mnaser): This is a list of all the Kubernetes versions which we've
#               released images for.  This list is used to determine which
#               images of Kubernetes we should publish to the registry.
VERSIONS = [
    "v1.23.13",
    "v1.23.17",
    "v1.24.7",
    "v1.24.15",
    "v1.25.3",
    "v1.25.11",
    "v1.26.2",
    "v1.26.6",
    "v1.26.11",
    "v1.27.3",
    "v1.27.8",
]


@click.command()
@click.option(
    "--repository",
    required=True,
    help="Target image repository",
)
@click.option(
    "--parallel",
    default=8,
    help="Number of parallel uploads",
)
@click.option(
    "--insecure",
    is_flag=True,
    help="Allow insecure connections to the registry.",
)
def main(repository, parallel, insecure):
    """
    Load images into a remote registry for `container_infra_prefix` usage.
    """
    crane_path = shutil.which("crane")

    if crane_path is None:
        raise click.UsageError(
            """Crane is not installed. Please install it before running this command:
             https://github.com/google/go-containerregistry/blob/main/cmd/crane/README.md#installation"""
        )

    # NOTE(mnaser): This list must be maintained manually because the image
    #               registry must be able to support a few different versions
    #               of Kubernetes since it is possible to have multiple
    #               clusters running different versions of Kubernetes at the
    #               same time.
    images = set(
        _get_all_kubeadm_images()
        + _get_calico_images()
        + _get_cloud_provider_images()
        + _get_infra_images()
    )

    with concurrent.futures.ThreadPoolExecutor(max_workers=parallel) as executor:
        future_to_image = {
            executor.submit(
                _mirror_image, image, repository, insecure, crane_path
            ): image
            for image in images
        }

        for future in concurrent.futures.as_completed(future_to_image):
            image = future_to_image[future]
            try:
                future.result()
            except Exception as e:
                click.echo(
                    f"Image upload failed for {image}: {e}",
                    err=True,
                )


def _mirror_image(image: str, repository: str, insecure: bool, crane_path: str):
    src = image
    dst = image_utils.get_image(image, repository)

    try:
        command = [crane_path]
        if insecure:
            command.append("--insecure")
        command += ["copy", "--platform", "linux/amd64", src, dst]

        subprocess.run(command, check=True)
    except subprocess.CalledProcessError:
        click.echo(
            "Image upload failed. Please ensure you're logged in via Crane.",
            err=True,
        )
        return


def _get_all_kubeadm_images():
    """
    Get the list of images that are used by Kubernetes by downloading "kubeadm"
    and running the "kubeadm config images list" command.
    """

    images = []
    for version in VERSIONS:
        images += _get_kubeadm_images(version)

    return images


@CACHE.memoize()
def _get_kubeadm_images(version: str):
    """
    Get the list of images that are used by Kubernetes by downloading "kubeadm"
    and running the "kubeadm config images list" command.
    """

    # Download kubeadm
    r = requests.get(f"https://dl.k8s.io/release/{version}/bin/linux/amd64/kubeadm")
    r.raise_for_status()

    # Write it to a temporary file
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(r.content)
        f.close()

        # Make it executable
        os.chmod(f.name, 0o755)

        # Run the command
        output = subprocess.check_output(
            [f.name, "config", "images", "list", "--kubernetes-version", version]
        )

        # Remove the temporary file
        os.unlink(f.name)

    # Parse the output
    return output.decode().replace("k8s.gcr.io", "registry.k8s.io").splitlines()


def _get_calico_images():
    return [
        # Calico 3.24.2
        "docker.io/calico/cni:v3.24.2",
        "docker.io/calico/kube-controllers:v3.24.2",
        "docker.io/calico/node:v3.24.2",
    ]


def _get_cloud_provider_images():
    return [
        # 1.24.6
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.24.6",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.24.6",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.24.6",
        # v1.25.3
        "docker.io/k8scloudprovider/cinder-csi-plugin:v1.25.3",
        "docker.io/k8scloudprovider/manila-csi-plugin:v1.25.3",
        "docker.io/k8scloudprovider/openstack-cloud-controller-manager:v1.25.3",
        # v1.25.6
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.25.6",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.25.6",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.25.6",
        # v1.26.3
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.26.3",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.26.3",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.26.3",
        # v1.27.2
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.27.2",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.27.2",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.27.2",
        # v1.28.0
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.28.0",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.28.0",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.28.0",
        # v1.29.0
        "registry.k8s.io/provider-os/k8s-keystone-auth:v1.29.0",
    ]


def _get_infra_images():
    return [
        "registry.k8s.io/sig-storage/csi-attacher:v3.4.0",
        "registry.k8s.io/sig-storage/csi-attacher:v4.2.0",
        "registry.k8s.io/sig-storage/csi-node-driver-registrar:v2.4.0",
        "registry.k8s.io/sig-storage/csi-node-driver-registrar:v2.5.1",
        "registry.k8s.io/sig-storage/csi-node-driver-registrar:v2.6.2",
        "registry.k8s.io/sig-storage/csi-node-driver-registrar:v2.6.3",
        "registry.k8s.io/sig-storage/csi-node-driver-registrar:v2.9.0",
        "registry.k8s.io/sig-storage/csi-provisioner:v3.0.0",
        "registry.k8s.io/sig-storage/csi-provisioner:v3.1.0",
        "registry.k8s.io/sig-storage/csi-provisioner:v3.3.0",
        "registry.k8s.io/sig-storage/csi-provisioner:v3.4.1",
        "registry.k8s.io/sig-storage/csi-resizer:v1.4.0",
        "registry.k8s.io/sig-storage/csi-resizer:v1.8.0",
        "registry.k8s.io/sig-storage/csi-snapshotter:v5.0.1",
        "registry.k8s.io/sig-storage/csi-snapshotter:v6.0.1",
        "registry.k8s.io/sig-storage/csi-snapshotter:v6.2.1",
        "registry.k8s.io/sig-storage/livenessprobe:v2.7.0",
        "registry.k8s.io/sig-storage/livenessprobe:v2.8.0",
        "registry.k8s.io/sig-storage/livenessprobe:v2.9.0",
        "registry.k8s.io/sig-storage/nfsplugin:v4.2.0",
    ]



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools
import os

import pkg_resources
import yaml

from magnum_cluster_api.cmd import image_loader


def _get_images_from_manifests(file: str):
    with open(file) as fd:
        data = fd.read()

    images = []
    for doc in yaml.safe_load_all(data):
        if doc["kind"] in ("DaemonSet", "Deployment", "StatefulSet"):
            for container in itertools.chain(
                doc["spec"]["template"]["spec"].get("initContainers", []),
                doc["spec"]["template"]["spec"]["containers"],
            ):
                images.append(container["image"])

    return set(images)


def test__get_calico_images():
    manifests_path = pkg_resources.resource_filename("magnum_cluster_api", "manifests")

    calico_path = os.path.join(manifests_path, "calico")
    for file in os.listdir(calico_path):
        assert _get_images_from_manifests(os.path.join(calico_path, file)) == set(
            image_loader._get_calico_images()
        )


def test__get_cloud_provider_images():
    manifests_path = pkg_resources.resource_filename("magnum_cluster_api", "manifests")
    ccm_path = os.path.join(
        manifests_path, "ccm/openstack-cloud-controller-manager-ds.yaml"
    )

    assert _get_images_from_manifests(ccm_path).issubset(
        image_loader._get_cloud_provider_images()
    )


def test__get_infra_images():
    manifests_path = pkg_resources.resource_filename("magnum_cluster_api", "manifests")

    for csi in ["cinder", "manila", "nfs"]:
        folder = os.path.join(manifests_path, f"{csi}-csi")

        for file in os.listdir(folder):
            path = os.path.join(folder, file)

            for image in _get_images_from_manifests(path):
                if image in ("registry.k8s.io/provider-os/manila-csi-plugin:latest",):
                    continue

                assert (
                    image in image_loader._get_infra_images()
                    or image in image_loader._get_cloud_provider_images()
                )



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from magnum.i18n import _
from oslo_config import cfg

auto_scaling_group = cfg.OptGroup(name="auto_scaling", title="Options for auto scaling")

capi_client_group = cfg.OptGroup(
    name="capi_client", title="Options for the Cluster API client"
)

manila_client_group = cfg.OptGroup(
    name="manila_client", title="Options for the Manila client"
)

proxy_group = cfg.OptGroup(name="proxy", title="Options for Cluster API proxy")


auto_scaling_opts = [
    cfg.StrOpt(
        "image_repository",
        default="registry.k8s.io/autoscaling",
        help="Image repository for the cluster auto-scaler.",
    ),
    cfg.StrOpt(
        "v1_22_image",
        default="$image_repository/cluster-autoscaler:v1.22.3",
        help="Image for the cluster auto-scaler for Kubernetes v1.22.",
    ),
    cfg.StrOpt(
        "v1_23_image",
        default="$image_repository/cluster-autoscaler:v1.23.1",
        help="Image for the cluster auto-scaler for Kubernetes v1.23.",
    ),
    cfg.StrOpt(
        "v1_24_image",
        default="$image_repository/cluster-autoscaler:v1.24.2",
        help="Image for the cluster auto-scaler for Kubernetes v1.24.",
    ),
    cfg.StrOpt(
        "v1_25_image",
        default="$image_repository/cluster-autoscaler:v1.25.2",
        help="Image for the cluster auto-scaler for Kubernetes v1.25.",
    ),
    cfg.StrOpt(
        "v1_26_image",
        default="$image_repository/cluster-autoscaler:v1.26.3",
        help="Image for the cluster auto-scaler for Kubernetes v1.26.",
    ),
    cfg.StrOpt(
        "v1_27_image",
        default="$image_repository/cluster-autoscaler:v1.27.2",
        help="Image for the cluster auto-scaler for Kubernetes v1.27.",
    ),
]


capi_client_opts = [
    cfg.StrOpt(
        "endpoint_type",
        default="publicURL",
        help=_(
            "Type of endpoint in Identity service catalog to use "
            "for communication with the OpenStack service."
        ),
    ),
]


manila_client_opts = [
    cfg.StrOpt(
        "region_name",
        help=_(
            "Region in Identity service catalog to use for "
            "communication with the OpenStack service."
        ),
    ),
    cfg.StrOpt(
        "endpoint_type",
        default="publicURL",
        help=_(
            "Type of endpoint in Identity service catalog to use "
            "for communication with the OpenStack service."
        ),
    ),
    cfg.StrOpt(
        "api_version",
        default="3",
        help=_("Version of Manila API to use in manilaclient."),
    ),
]


proxy_opts = [
    cfg.StrOpt(
        "haproxy_pid_path",
        default="/var/run/haproxy.pid",
        help=_("Path to HAProxy PID file."),
    ),
]


common_security_opts = [
    cfg.StrOpt("ca_file", help=_("Optional CA cert file to use in SSL connections.")),
    cfg.StrOpt("cert_file", help=_("Optional PEM-formatted certificate chain file.")),
    cfg.StrOpt(
        "key_file",
        help=_("Optional PEM-formatted file that contains the " "private key."),
    ),
    cfg.BoolOpt(
        "insecure",
        default=False,
        help=_("If set, then the server's certificate will not " "be verified."),
    ),
]

CONF = cfg.CONF
CONF.register_group(auto_scaling_group)
CONF.register_group(capi_client_group)
CONF.register_group(manila_client_group)
CONF.register_group(proxy_group)
CONF.register_opts(auto_scaling_opts, group=auto_scaling_group)
CONF.register_opts(capi_client_opts, group=capi_client_group)
CONF.register_opts(common_security_opts, group=capi_client_group)
CONF.register_opts(manila_client_opts, group=manila_client_group)
CONF.register_opts(common_security_opts, group=manila_client_group)
CONF.register_opts(proxy_opts, group=proxy_group)



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import semver

from magnum_cluster_api import conf

CONF = conf.CONF

PAUSE = "registry.k8s.io/pause:3.9"


def get_cluster_autoscaler_image(version: str):
    version = semver.VersionInfo.parse(version[1:])
    config_option = f"v{version.major}_{version.minor}_image"

    if hasattr(CONF.auto_scaling, config_option):
        return getattr(CONF.auto_scaling, config_option)

    raise ValueError(
        f"Unsupported Kubernetes version: {version}. "
        "Please specify a supported version in the cluster template."
    )



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import glob
import os
import uuid

import pkg_resources
import pytest
import yaml

from magnum_cluster_api import image_utils


@pytest.mark.parametrize(
    "glob_path",
    [
        "calico/*.yaml",
        "ccm/*.yaml",
        "csi/*.yaml",
    ],
)
def test_update_manifest_images(glob_path):
    manifests_path = pkg_resources.resource_filename("magnum_cluster_api", "manifests")
    repository = "quay.io/test"

    cluster_uuid = str(uuid.uuid4())

    for manifest_file in glob.glob(os.path.join(manifests_path, glob_path)):
        data = image_utils.update_manifest_images(
            cluster_uuid,
            manifest_file,
            repository=repository,
        )

        for doc in yaml.safe_load_all(data):
            if doc["kind"] in ("DaemonSet", "Deployment"):
                for init_container in doc["spec"]["template"]["spec"].get(
                    "initContainers", []
                ):
                    assert init_container["image"].startswith(repository)
                for container in doc["spec"]["template"]["spec"]["containers"]:
                    assert container["image"].startswith(repository)



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import pytest
from magnum.objects import fields
from magnum.tests.unit.objects import utils

from magnum_cluster_api import resources


def test_generate_machine_deployments_for_cluster_with_deleting_node_group(
    context, mocker
):
    cluster_template = mocker.Mock()
    cluster_template.labels = {"kube_tag": "v1.26.2"}

    cluster = mocker.Mock()
    cluster.cluster_template = cluster_template
    cluster.labels = {}
    cluster.nodegroups = [
        mocker.Mock(
            name="creating-worker",
            status=fields.ClusterStatus.CREATE_IN_PROGRESS,
            labels={},
        ),
        mocker.Mock(
            name="created-worker",
            status=fields.ClusterStatus.CREATE_COMPLETE,
            labels={},
        ),
        mocker.Mock(
            name="deleting-worker",
            status=fields.ClusterStatus.DELETE_IN_PROGRESS,
            labels={},
        ),
        mocker.Mock(
            name="deleted-worker",
            status=fields.ClusterStatus.DELETE_COMPLETE,
            labels={},
        ),
    ]

    cluster_get_by_uuid = mocker.patch("magnum.objects.Cluster.get_by_uuid")
    cluster_get_by_uuid.return_value = cluster

    mock_get_default_boot_volume_type = mocker.patch(
        "magnum_cluster_api.integrations.cinder.get_default_boot_volume_type"
    )
    mock_get_default_boot_volume_type.return_value = "foo"

    mock_get_image_uuid = mocker.patch("magnum_cluster_api.utils.get_image_uuid")
    mock_get_image_uuid.return_value = "foo"

    mds = resources.generate_machine_deployments_for_cluster(
        context,
        cluster,
    )

    assert len(mds) == 2


@pytest.mark.parametrize(
    "auto_scaling_enabled",
    [True, False, None],
    ids=lambda x: f"auto_scaling_enabled={x}",
)
@pytest.mark.parametrize(
    "auto_healing_enabled",
    [True, False, None],
    ids=lambda x: f"auto_healing_enabled={x}",
)
class TestExistingMutateMachineDeployment:
    @pytest.fixture(autouse=True)
    def setup(self, auto_scaling_enabled, auto_healing_enabled, context):
        self.cluster = utils.get_test_cluster(context, labels={})
        if auto_scaling_enabled is not None:
            self.cluster.labels["auto_scaling_enabled"] = str(auto_scaling_enabled)

        if auto_healing_enabled is not None:
            self.cluster.labels["auto_healing_enabled"] = str(auto_healing_enabled)

        self.node_group = utils.get_test_nodegroup(context, labels={})
        if auto_scaling_enabled is not None:
            self.node_group.min_node_count = 1
            self.node_group.max_node_count = 3

    def _assert_no_mutations(self, md):
        assert md["name"] == self.node_group.name
        assert "class" not in md

    def _assert_common_machine_deployment_values(self, md):
        assert md["name"] == self.node_group.name
        assert md["metadata"]["labels"] == {
            f"node-role.kubernetes.io/{self.node_group.role}": "",
            "node.cluster.x-k8s.io/nodegroup": self.node_group.name,
        }
        assert (
            md["nodeVolumeDetachTimeout"]
            == resources.CLUSTER_CLASS_NODE_VOLUME_DETACH_TIMEOUT
        )

    def test_mutate_machine_deployment(self, context, auto_scaling_enabled):
        md = resources.mutate_machine_deployment(
            context,
            self.cluster,
            self.node_group,
            {
                "name": self.node_group.name,
            },
        )

        self._assert_common_machine_deployment_values(md)
        self._assert_no_mutations(md)

        if auto_scaling_enabled:
            assert md["replicas"] is None
            assert md["metadata"]["annotations"][
                resources.AUTOSCALE_ANNOTATION_MIN
            ] == str(self.node_group.min_node_count)
            assert md["metadata"]["annotations"][
                resources.AUTOSCALE_ANNOTATION_MAX
            ] == str(self.node_group.max_node_count)
        else:
            assert md["replicas"] == self.node_group.node_count
            assert md["metadata"]["annotations"] == {}



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import configparser
import copy
import io

import pykube
import yaml
from magnum import objects as magnum_objects
from oslo_serialization import base64
from tenacity import Retrying, retry_if_result, stop_after_delay, wait_fixed

from magnum_cluster_api import exceptions


class NamespacedAPIObject(pykube.objects.NamespacedAPIObject):
    @property
    def events(self):
        return pykube.Event.objects(self.api, namespace=self.namespace).filter(
            field_selector={
                "involvedObject.name": self.name,
                "involvedObject.apiVersion": self.version,
                "involvedObject.kind": self.kind,
            },
        )

    @property
    def observed_generation(self):
        return self.obj.get("status", {}).get("observedGeneration")

    def wait_for_observed_generation_changed(
        self,
        existing_observed_generation: int = 0,
        timeout: int = 10,
        interval: int = 1,
    ):
        if existing_observed_generation == 0:
            existing_observed_generation = self.observed_generation

        for attempt in Retrying(
            retry=(
                retry_if_result(lambda g: g == existing_observed_generation)
                | retry_if_result(lambda g: g is None)
            ),
            stop=stop_after_delay(timeout),
            wait=wait_fixed(interval),
        ):
            with attempt:
                self.reload()
            if not attempt.retry_state.outcome.failed:
                attempt.retry_state.set_result(self.observed_generation)


class EndpointSlice(NamespacedAPIObject):
    version = "discovery.k8s.io/v1"
    endpoint = "endpointslices"
    kind = "EndpointSlice"


class ClusterResourceSet(NamespacedAPIObject):
    version = "addons.cluster.x-k8s.io/v1beta1"
    endpoint = "clusterresourcesets"
    kind = "ClusterResourceSet"


class OpenStackMachineTemplate(NamespacedAPIObject):
    version = "infrastructure.cluster.x-k8s.io/v1alpha7"
    endpoint = "openstackmachinetemplates"
    kind = "OpenStackMachineTemplate"


class KubeadmConfigTemplate(NamespacedAPIObject):
    version = "bootstrap.cluster.x-k8s.io/v1beta1"
    endpoint = "kubeadmconfigtemplates"
    kind = "KubeadmConfigTemplate"


class KubeadmControlPlane(NamespacedAPIObject):
    version = "controlplane.cluster.x-k8s.io/v1beta1"
    endpoint = "kubeadmcontrolplanes"
    kind = "KubeadmControlPlane"


class KubeadmControlPlaneTemplate(NamespacedAPIObject):
    version = "controlplane.cluster.x-k8s.io/v1beta1"
    endpoint = "kubeadmcontrolplanetemplates"
    kind = "KubeadmControlPlaneTemplate"


class MachineDeployment(NamespacedAPIObject):
    version = "cluster.x-k8s.io/v1beta1"
    endpoint = "machinedeployments"
    kind = "MachineDeployment"

    @classmethod
    def for_node_group(
        cls,
        api: pykube.HTTPClient,
        cluster: magnum_objects.Cluster,
        node_group: magnum_objects.NodeGroup,
    ):
        mds = cls.objects(api, namespace="magnum-system").filter(
            selector={
                "cluster.x-k8s.io/cluster-name": cluster.stack_id,
                "topology.cluster.x-k8s.io/deployment-name": node_group.name,
            },
        )
        if len(mds) != 1:
            raise exceptions.MachineDeploymentNotFound(name=node_group.name)
        return list(mds)[0]

    @classmethod
    def for_node_group_or_none(
        cls,
        api: pykube.HTTPClient,
        cluster: magnum_objects.Cluster,
        node_group: magnum_objects.NodeGroup,
    ):
        try:
            return cls.for_node_group(api, cluster, node_group)
        except exceptions.MachineDeploymentNotFound:
            return None

    def equals_spec(self, spec: dict) -> bool:
        expected_annotations = spec["metadata"].get("annotations")
        current_annotations = self.obj["spec"]["template"]["metadata"].get(
            "annotations"
        )

        # NOTE(mnaser): If we have any annotations, that means that autoscaling is
        #               enabled and we should not compare the replicas.
        if expected_annotations:
            return expected_annotations == current_annotations

        expected_replicas = spec.get("replicas")
        current_replicas = self.obj["spec"].get("replicas")

        return expected_replicas == current_replicas


class Machine(NamespacedAPIObject):
    version = "cluster.x-k8s.io/v1beta1"
    endpoint = "machines"
    kind = "Machine"


class OpenStackClusterTemplate(NamespacedAPIObject):
    version = "infrastructure.cluster.x-k8s.io/v1alpha7"
    endpoint = "openstackclustertemplates"
    kind = "OpenStackClusterTemplate"


class OpenStackCluster(NamespacedAPIObject):
    version = "infrastructure.cluster.x-k8s.io/v1alpha7"
    endpoint = "openstackclusters"
    kind = "OpenStackCluster"

    @property
    def identity_ref(self):
        return self.obj["spec"]["identityRef"]

    @property
    def identity_ref_secret(self) -> pykube.Secret:
        return pykube.Secret.objects(self.api, namespace=self.namespace).get(
            name=self.identity_ref["name"]
        )

    @property
    def clouds_yaml(self):
        return yaml.safe_load(
            base64.decode_as_text(self.identity_ref_secret.obj["data"]["clouds.yaml"])
        )

    @property
    def cloud_config(self):
        return self.clouds_yaml["clouds"]["default"]

    @property
    def floating_network_id(self):
        try:
            return self.obj["status"]["externalNetwork"]["id"]
        except KeyError:
            raise exceptions.OpenStackClusterExternalNetworkNotReady()

    @property
    def network_id(self):
        try:
            return self.obj["status"]["network"]["id"]
        except KeyError:
            raise exceptions.OpenStackClusterNetworkNotReady()

    @property
    def subnet_id(self):
        try:
            return self.obj["status"]["network"]["subnet"]["id"]
        except KeyError:
            raise exceptions.OpenStackClusterSubnetNotReady()

    @property
    def cloud_controller_manager_config(self):
        config = configparser.ConfigParser()
        config["Global"] = {
            "auth-url": self.cloud_config["auth"]["auth_url"],
            "region": self.cloud_config["region_name"],
            "application-credential-id": self.cloud_config["auth"][
                "application_credential_id"
            ],
            "application-credential-secret": self.cloud_config["auth"][
                "application_credential_secret"
            ],
            "tls-insecure": "false" if self.cloud_config["verify"] else "true",
        }
        config["LoadBalancer"] = {
            "floating-network-id": self.floating_network_id,
            "network-id": self.network_id,
            "subnet-id": self.subnet_id,
        }

        fd = io.StringIO()
        config.write(fd)

        return fd.getvalue()


class ClusterClass(NamespacedAPIObject):
    version = "cluster.x-k8s.io/v1beta1"
    endpoint = "clusterclasses"
    kind = "ClusterClass"


class Cluster(NamespacedAPIObject):
    version = "cluster.x-k8s.io/v1beta1"
    endpoint = "clusters"
    kind = "Cluster"

    @classmethod
    def for_magnum_cluster(
        cls, api: pykube.HTTPClient, cluster: magnum_objects.Cluster
    ) -> "Cluster":
        return cls.objects(api, namespace="magnum-system").get(name=cluster.stack_id)

    @classmethod
    def for_magnum_cluster_or_none(
        cls, api: pykube.HTTPClient, cluster: magnum_objects.Cluster
    ) -> "Cluster":
        return cls.objects(api, namespace="magnum-system").get_or_none(
            name=cluster.stack_id
        )

    def get_machine_deployment_index(self, name: str) -> int:
        for i, machine_deployment in enumerate(
            self.obj["spec"]["topology"]["workers"]["machineDeployments"]
        ):
            if machine_deployment["name"] == name:
                return i

        raise exceptions.MachineDeploymentNotFound(name=name)

    def get_machine_deployment_spec(self, name: str) -> dict:
        return copy.deepcopy(
            self.obj["spec"]["topology"]["workers"]["machineDeployments"][
                self.get_machine_deployment_index(name)
            ]
        )

    def set_machine_deployment_spec(self, name: str, spec: dict):
        self.obj["spec"]["topology"]["workers"]["machineDeployments"][
            self.get_machine_deployment_index(name)
        ] = spec

    @property
    def openstack_cluster(self):
        filtered_clusters = (
            OpenStackCluster.objects(self.api, namespace=self.namespace)
            .filter(selector={"cluster.x-k8s.io/cluster-name": self.name})
            .all()
        )

        if len(filtered_clusters) == 0:
            raise exceptions.OpenStackClusterNotCreated()

        return list(filtered_clusters)[0]


class OpenStackMachine(NamespacedAPIObject):
    version = "infrastructure.cluster.x-k8s.io/v1alpha7"
    endpoint = "openstackmachines"
    kind = "OpenStackMachine"


class StorageClass(pykube.objects.APIObject):
    version = "storage.k8s.io/v1"
    endpoint = "storageclasses"
    kind = "StorageClass"



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import pytest

from magnum_cluster_api import objects


@pytest.fixture
def cluster(
    context,
    cluster_obj,
    ubuntu_driver,
    mock_validate_cluster,
    mock_osc,
    mock_certificates,
):
    try:
        ubuntu_driver.create_cluster(context, cluster_obj, 60)

        cluster_resource = objects.Cluster.for_magnum_cluster(
            ubuntu_driver.k8s_api, cluster_obj
        )
        cluster_resource.wait_for_observed_generation_changed(
            existing_observed_generation=1
        )

        cluster_obj.save.assert_called_once()
        cluster_obj.save.reset_mock()

        yield cluster_obj
    finally:
        ubuntu_driver.delete_cluster(context, cluster_obj)
